@@ -5,6 +5,7 @@ import logging
 import os
 import re
 import shlex
+import sys
 import tarfile
 import tempfile
 from abc import ABCMeta, abstractmethod
@@ -12,8 +13,15 @@ from enum import Enum, unique
 from pathlib import Path
 from typing import Dict, List, Optional, Tuple, Union
 
+if sys.version_info >= (3, 8):
+    from typing import Protocol
+else:
+    from typing_extensions import Protocol
+
 from localstack import config
-from localstack.utils.common import TMP_FILES, HashableList, rm_rf, save_file, short_uid
+from localstack.utils.collections import HashableList
+from localstack.utils.files import TMP_FILES, rm_rf, save_file
+from localstack.utils.strings import short_uid
 
 LOG = logging.getLogger(__name__)
 
@@ -75,28 +83,55 @@ class AccessDenied(ContainerException):
         self.object_name = object_name
 
 
-class PortMappings(object):
+class CancellableStream(Protocol):
+    """Describes a generator that can be closed. Borrowed from ``docker.types.daemon``."""
+
+    def __iter__(self):
+        raise NotImplementedError
+
+    def __next__(self):
+        raise NotImplementedError
+
+    def close(self):
+        raise NotImplementedError
+
+
+# defines the type for port mappings (source->target port range)
+PortRange = Union[List, HashableList]
+
+
+class PortMappings:
     """Maps source to target port ranges for Docker port mappings."""
 
-    def __init__(self, bind_host=None):
+    # bind host to be used for defining port mappings
+    bind_host: str
+    # maps `from` port range to `to` port range for port mappings
+    mappings: Dict[PortRange, List]
+
+    def __init__(self, bind_host: str = None):
         self.bind_host = bind_host if bind_host else ""
         self.mappings = {}
 
-    def add(self, port, mapped=None, protocol="tcp"):
+    def add(
+        self,
+        port: Union[int, PortRange],
+        mapped: Union[int, PortRange] = None,
+        protocol: str = "tcp",
+    ):
         mapped = mapped or port
-        if isinstance(port, list):
+        if isinstance(port, PortRange):
             for i in range(port[1] - port[0] + 1):
-                if isinstance(mapped, list):
+                if isinstance(mapped, PortRange):
                     self.add(port[0] + i, mapped[0] + i)
                 else:
                     self.add(port[0] + i, mapped)
             return
         if port is None or int(port) <= 0:
-            raise Exception("Unable to add mapping for invalid port: %s" % port)
+            raise Exception(f"Unable to add mapping for invalid port: {port}")
         if self.contains(port):
             return
         bisected_host_port = None
-        for from_range, to_range in self.mappings.items():
+        for from_range, to_range in dict(self.mappings).items():
             if not self.in_expanded_range(port, from_range):
                 continue
             if not self.in_expanded_range(mapped, to_range):
@@ -105,20 +140,20 @@ class PortMappings(object):
             to_range_len = to_range[1] - to_range[0]
             is_uniform = from_range_len == to_range_len
             if is_uniform:
-                self.expand_range(port, from_range)
+                self.expand_range(port, from_range, remap=True)
                 self.expand_range(mapped, to_range)
             else:
                 if not self.in_range(mapped, to_range):
                     continue
                 # extending a 1 to 1 mapping to be many to 1
                 elif from_range_len == 1:
-                    self.expand_range(port, from_range)
+                    self.expand_range(port, from_range, remap=True)
                 # splitting a uniform mapping
                 else:
                     bisected_port_index = mapped - to_range[0]
                     bisected_host_port = from_range[0] + bisected_port_index
                     self.bisect_range(mapped, to_range)
-                    self.bisect_range(bisected_host_port, from_range)
+                    self.bisect_range(bisected_host_port, from_range, remap=True)
                     break
             return
         protocol = str(protocol or "tcp").lower()
@@ -179,39 +214,51 @@ class PortMappings(object):
         items = [item for k, v in self.mappings.items() for item in entry(k, v)]
         return dict(items)
 
-    def contains(self, port):
+    def contains(self, port: int) -> bool:
         for from_range, to_range in self.mappings.items():
             if self.in_range(port, from_range):
                 return True
 
-    def in_range(self, port, range):
+    def in_range(self, port: int, range: PortRange) -> bool:
         return port >= range[0] and port <= range[1]
 
-    def in_expanded_range(self, port, range):
+    def in_expanded_range(self, port: int, range: PortRange):
         return port >= range[0] - 1 and port <= range[1] + 1
 
-    def expand_range(self, port, range):
+    def expand_range(self, port: int, range: PortRange, remap: bool = False):
+        """
+        Expand the given port range by the given port. If remap==True, put the updated range into self.mappings
+        """
         if self.in_range(port, range):
             return
+        new_range = list(range) if remap else range
         if port == range[0] - 1:
-            range[0] = port
+            new_range[0] = port
         elif port == range[1] + 1:
-            range[1] = port
+            new_range[1] = port
         else:
-            raise Exception("Unable to add port %s to existing range %s" % (port, range))
+            raise Exception(f"Unable to add port {port} to existing range {range}")
+        if remap:
+            self._remap_range(range, new_range)
 
-    """Bisect a port range, at the provided port
-        This is needed in some cases when adding a non-uniform host to port mapping
-        adjacent to an existing port range
-    """
-
-    def bisect_range(self, port, range):
+    def bisect_range(self, port: int, range: PortRange, remap: bool = False):
+        """
+        Bisect a port range, at the provided port. This is needed in some cases when adding a
+        non-uniform host to port mapping adjacent to an existing port range.
+        If remap==True, put the updated range into self.mappings
+        """
         if not self.in_range(port, range):
             return
+        new_range = list(range) if remap else range
         if port == range[0]:
-            range[0] = port + 1
+            new_range[0] = port + 1
         else:
-            range[1] = port - 1
+            new_range[1] = port - 1
+        if remap:
+            self._remap_range(range, new_range)
+
+    def _remap_range(self, old_key: PortRange, new_key: PortRange):
+        self.mappings[HashableList(new_key)] = self.mappings.pop(old_key)
 
     def __repr__(self):
         return f"<PortMappings: {self.to_dict()}>"
@@ -290,6 +337,7 @@ class ContainerConfiguration:
     user: Optional[str] = None
     cap_add: Optional[List[str]] = None
     cap_drop: Optional[List[str]] = None
+    security_opt: Optional[List[str]] = None
     network: Optional[str] = None
     dns: Optional[str] = None
     workdir: Optional[str] = None
@@ -405,12 +453,12 @@ class ContainerClient(metaclass=ABCMeta):
 
     @abstractmethod
     def pull_image(self, docker_image: str) -> None:
-        """Pulls a image with a given name from a Docker registry"""
+        """Pulls an image with a given name from a Docker registry"""
         pass
 
     @abstractmethod
     def push_image(self, docker_image: str) -> None:
-        """Pushes a image with a given name to a Docker registry"""
+        """Pushes an image with a given name to a Docker registry"""
         pass
 
     @abstractmethod
@@ -447,6 +495,11 @@ class ContainerClient(metaclass=ABCMeta):
         """Get all logs of a given container"""
         pass
 
+    @abstractmethod
+    def stream_container_logs(self, container_name_or_id: str) -> CancellableStream:
+        """Returns a blocking generator you can iterate over to retrieve log output as it happens."""
+        pass
+
     @abstractmethod
     def inspect_container(self, container_name_or_id: str) -> Dict[str, Union[Dict, str]]:
         """Get detailed attributes of an container.
@@ -551,6 +604,34 @@ class ContainerClient(metaclass=ABCMeta):
         """
         pass
 
+    def create_container_from_config(self, container_config: ContainerConfiguration) -> str:
+        """
+        Similar to create_container, but allows passing the whole ContainerConfiguration
+        :param container_config: ContainerConfiguration how to start the container
+        :return: Container ID
+        """
+        return self.create_container(
+            image_name=container_config.image_name,
+            name=container_config.name,
+            entrypoint=container_config.entrypoint,
+            remove=container_config.remove,
+            interactive=container_config.interactive,
+            tty=container_config.tty,
+            detach=container_config.detach,
+            command=container_config.command,
+            mount_volumes=container_config.volumes,
+            ports=container_config.ports,
+            env_vars=container_config.env_vars,
+            user=container_config.user,
+            cap_add=container_config.cap_add,
+            cap_drop=container_config.cap_drop,
+            security_opt=container_config.security_opt,
+            network=container_config.network,
+            dns=container_config.dns,
+            additional_flags=container_config.additional_flags,
+            workdir=container_config.workdir,
+        )
+
     @abstractmethod
     def create_container(
         self,
@@ -569,6 +650,7 @@ class ContainerClient(metaclass=ABCMeta):
         user: Optional[str] = None,
         cap_add: Optional[List[str]] = None,
         cap_drop: Optional[List[str]] = None,
+        security_opt: Optional[List[str]] = None,
         network: Optional[str] = None,
         dns: Optional[str] = None,
         additional_flags: Optional[str] = None,
@@ -599,6 +681,7 @@ class ContainerClient(metaclass=ABCMeta):
         user: Optional[str] = None,
         cap_add: Optional[List[str]] = None,
         cap_drop: Optional[List[str]] = None,
+        security_opt: Optional[List[str]] = None,
         network: Optional[str] = None,
         dns: Optional[str] = None,
         additional_flags: Optional[str] = None,
