@@ -10,6 +10,7 @@ from typing import Dict, List, Optional, Tuple, Union
 from localstack import config
 from localstack.utils.container_utils.container_client import (
     AccessDenied,
+    CancellableStream,
     ContainerClient,
     ContainerException,
     DockerContainerStatus,
@@ -28,6 +29,26 @@ from localstack.utils.strings import to_str
 LOG = logging.getLogger(__name__)
 
 
+class CancellableProcessStream(CancellableStream):
+    process: subprocess.Popen
+
+    def __init__(self, process: subprocess.Popen) -> None:
+        super().__init__()
+        self.process = process
+
+    def __iter__(self):
+        return self
+
+    def __next__(self):
+        line = self.process.stdout.readline()
+        if not line:
+            raise StopIteration
+        return line
+
+    def close(self):
+        return self.process.terminate()
+
+
 class CmdDockerClient(ContainerClient):
     """Class for managing docker containers using the command line executable"""
 
@@ -75,7 +96,7 @@ class CmdDockerClient(ContainerClient):
             else:
                 raise ContainerException(
                     "Docker process returned with errorcode %s" % e.returncode, e.stdout, e.stderr
-                )
+                ) from e
 
     def pause_container(self, container_name: str) -> None:
         cmd = self._docker_cmd()
@@ -89,7 +110,7 @@ class CmdDockerClient(ContainerClient):
             else:
                 raise ContainerException(
                     "Docker process returned with errorcode %s" % e.returncode, e.stdout, e.stderr
-                )
+                ) from e
 
     def unpause_container(self, container_name: str) -> None:
         cmd = self._docker_cmd()
@@ -103,7 +124,7 @@ class CmdDockerClient(ContainerClient):
             else:
                 raise ContainerException(
                     "Docker process returned with errorcode %s" % e.returncode, e.stdout, e.stderr
-                )
+                ) from e
 
     def remove_image(self, image: str, force: bool = True) -> None:
         cmd = self._docker_cmd()
@@ -119,7 +140,7 @@ class CmdDockerClient(ContainerClient):
             else:
                 raise ContainerException(
                     "Docker process returned with errorcode %s" % e.returncode, e.stdout, e.stderr
-                )
+                ) from e
 
     def commit(
         self,
@@ -140,7 +161,7 @@ class CmdDockerClient(ContainerClient):
             else:
                 raise ContainerException(
                     "Docker process returned with errorcode %s" % e.returncode, e.stdout, e.stderr
-                )
+                ) from e
 
     def remove_container(self, container_name: str, force=True, check_existence=False) -> None:
         if check_existence and container_name not in self.get_running_container_names():
@@ -158,7 +179,7 @@ class CmdDockerClient(ContainerClient):
             else:
                 raise ContainerException(
                     "Docker process returned with errorcode %s" % e.returncode, e.stdout, e.stderr
-                )
+                ) from e
 
     def list_containers(self, filter: Union[List[str], str, None] = None, all=True) -> List[dict]:
         filter = [filter] if isinstance(filter, str) else filter
@@ -177,7 +198,7 @@ class CmdDockerClient(ContainerClient):
         except subprocess.CalledProcessError as e:
             raise ContainerException(
                 "Docker process returned with errorcode %s" % e.returncode, e.stdout, e.stderr
-            )
+            ) from e
         container_list = []
         if cmd_result:
             container_list = [json.loads(line) for line in cmd_result.splitlines()]
@@ -207,7 +228,7 @@ class CmdDockerClient(ContainerClient):
                 raise NoSuchContainer(container_name)
             raise ContainerException(
                 f"Docker process returned with errorcode {e.returncode}", e.stdout, e.stderr
-            )
+            ) from e
 
     def copy_from_container(
         self, container_name: str, local_path: str, container_path: str
@@ -222,7 +243,7 @@ class CmdDockerClient(ContainerClient):
                 raise NoSuchContainer(container_name)
             raise ContainerException(
                 "Docker process returned with errorcode %s" % e.returncode, e.stdout, e.stderr
-            )
+            ) from e
 
     def pull_image(self, docker_image: str) -> None:
         cmd = self._docker_cmd()
@@ -235,7 +256,7 @@ class CmdDockerClient(ContainerClient):
                 raise NoSuchImage(docker_image)
             raise ContainerException(
                 "Docker process returned with errorcode %s" % e.returncode, e.stdout, e.stderr
-            )
+            ) from e
 
     def push_image(self, docker_image: str) -> None:
         cmd = self._docker_cmd()
@@ -308,7 +329,19 @@ class CmdDockerClient(ContainerClient):
             else:
                 raise ContainerException(
                     "Docker process returned with errorcode %s" % e.returncode, e.stdout, e.stderr
-                )
+                ) from e
+
+    def stream_container_logs(self, container_name_or_id: str) -> CancellableStream:
+        self.inspect_container(container_name_or_id)  # guard to check whether container is there
+
+        cmd = self._docker_cmd()
+        cmd += ["logs", container_name_or_id, "--follow"]
+
+        process: subprocess.Popen = run(
+            cmd, asynchronous=True, outfile=subprocess.PIPE, stderr=subprocess.PIPE
+        )
+
+        return CancellableProcessStream(process)
 
     def _inspect_object(self, object_name_or_id: str) -> Dict[str, Union[Dict, str]]:
         cmd = self._docker_cmd()
@@ -321,7 +354,7 @@ class CmdDockerClient(ContainerClient):
             else:
                 raise ContainerException(
                     "Docker process returned with errorcode %s" % e.returncode, e.stdout, e.stderr
-                )
+                ) from e
         image_data = json.loads(cmd_result.strip())
         return image_data
 
@@ -371,7 +404,7 @@ class CmdDockerClient(ContainerClient):
             else:
                 raise ContainerException(
                     "Docker process returned with errorcode %s" % e.returncode, e.stdout, e.stderr
-                )
+                ) from e
 
     def disconnect_container_from_network(
         self, network_name: str, container_name_or_id: str
@@ -391,7 +424,7 @@ class CmdDockerClient(ContainerClient):
             else:
                 raise ContainerException(
                     "Docker process returned with errorcode %s" % e.returncode, e.stdout, e.stderr
-                )
+                ) from e
 
     def get_container_ip(self, container_name_or_id: str) -> str:
         cmd = self._docker_cmd()
@@ -410,13 +443,13 @@ class CmdDockerClient(ContainerClient):
             else:
                 raise ContainerException(
                     "Docker process returned with errorcode %s" % e.returncode, e.stdout, e.stderr
-                )
+                ) from e
 
     def has_docker(self) -> bool:
         try:
             run(self._docker_cmd() + ["ps"])
             return True
-        except subprocess.CalledProcessError:
+        except (subprocess.CalledProcessError, FileNotFoundError):
             return False
 
     def create_container(self, image_name: str, **kwargs) -> str:
@@ -432,7 +465,7 @@ class CmdDockerClient(ContainerClient):
                 raise NoSuchImage(image_name, stdout=e.stdout, stderr=e.stderr)
             raise ContainerException(
                 "Docker process returned with errorcode %s" % e.returncode, e.stdout, e.stderr
-            )
+            ) from e
         finally:
             Util.rm_env_vars_file(env_file)
 
@@ -525,7 +558,7 @@ class CmdDockerClient(ContainerClient):
                 raise NoSuchContainer(container_name, stdout=e.stdout, stderr=e.stderr)
             raise ContainerException(
                 "Docker process returned with errorcode %s" % e.returncode, e.stdout, e.stderr
-            )
+            ) from e
 
     def _build_run_create_cmd(
         self,
@@ -545,6 +578,7 @@ class CmdDockerClient(ContainerClient):
         user: Optional[str] = None,
         cap_add: Optional[List[str]] = None,
         cap_drop: Optional[List[str]] = None,
+        security_opt: Optional[List[str]] = None,
         network: Optional[str] = None,
         dns: Optional[str] = None,
         additional_flags: Optional[str] = None,
@@ -581,6 +615,10 @@ class CmdDockerClient(ContainerClient):
             cmd += list(itertools.chain.from_iterable(["--cap-add", cap] for cap in cap_add))
         if cap_drop:
             cmd += list(itertools.chain.from_iterable(["--cap-drop", cap] for cap in cap_drop))
+        if security_opt:
+            cmd += list(
+                itertools.chain.from_iterable(["--security-opt", opt] for opt in security_opt)
+            )
         if network:
             cmd += ["--network", network]
         if dns:
